@@ -1,7 +1,6 @@
 from statefun_tasks.context import TaskContext
 from statefun_tasks.messages_pb2 import ChildPipeline, TaskRequest, TaskResult, TaskException, Pipeline, TaskActionRequest, TaskAction, TaskStatus
-from statefun_tasks.serialisation import DefaultSerialiser
-from statefun_tasks.storage import StorageBackend
+from statefun_tasks.serialisation import DefaultSerialiser, pack_any
 from statefun_tasks.types import Task, Group, TaskCancelledException, PipelineInProgress
 from statefun_tasks.type_helpers import _create_task_exception
 from statefun_tasks.pipeline_impl.handlers import BeginPipelineHandler, ContinuePipelineHandler, EndPipelineHandler, CancelPipelineHandler
@@ -13,24 +12,26 @@ from typing import Union
 
 
 class _Pipeline(object):
-    __slots__ = ('_pipeline', '_serialiser', '_is_fruitful', '_events', '_storage', '_handlers', '_graph', '_submitter')
+    __slots__ = ('_pipeline', '_serialiser', '_is_fruitful', '_events', '_inline', '_initial_args', '_initial_state', '_handlers', '_graph', '_submitter')
 
-    def __init__(self, pipeline: list, serialiser=None, is_fruitful=True, events: EventHandlers=None, storage: StorageBackend=None):
+    def __init__(self, pipeline: list, serialiser=None, is_fruitful=True, inline=False, initial_args=None, initial_state=None, events: EventHandlers=None):
         self._pipeline = pipeline
         self._serialiser = serialiser or DefaultSerialiser()
         self._is_fruitful = is_fruitful
+        self._inline = inline
+        self._initial_args = initial_args
+        self._initial_state = initial_state
         self._events = events or EventHandlers()
-        self._storage = storage
 
         self._handlers = [
-            BeginPipelineHandler(self._pipeline, self._serialiser, self._storage),
-            ContinuePipelineHandler(self._pipeline, self._serialiser, self._storage),
-            CancelPipelineHandler(self._pipeline, self._serialiser, self._storage),
-            EndPipelineHandler(self._pipeline, self._serialiser, self._storage)
+            BeginPipelineHandler(self._pipeline, self._serialiser),
+            ContinuePipelineHandler(self._pipeline, self._serialiser),
+            CancelPipelineHandler(self._pipeline, self._serialiser),
+            EndPipelineHandler(self._pipeline, self._serialiser)
         ]
 
         self._graph = PipelineGraph(self._pipeline)
-        self._submitter = DeferredTaskSubmitter(self._graph, self._serialiser, self._storage)
+        self._submitter = DeferredTaskSubmitter(self._graph, self._serialiser)
 
     @property
     def events(self) -> EventHandlers:
@@ -43,18 +44,35 @@ class _Pipeline(object):
     def is_fruitful(self):
         return self._is_fruitful
 
+    @property
+    def is_inline(self):
+        return self._inline
+
     def to_proto(self) -> Pipeline:
-        pipeline = Pipeline(entries=[p.to_proto(self._serialiser) for p in self._pipeline])
+        pipeline = Pipeline(
+            entries=[p.to_proto(self._serialiser) for p in self._pipeline], 
+            inline=self._inline)
+
+        if self._initial_args is not None:
+            pipeline.initial_args.CopyFrom(pack_any(self._serialiser.to_proto(self._initial_args)))
+        
+        if self._initial_state is not None:
+            pipeline.initial_state.CopyFrom(pack_any(self._serialiser.to_proto(self._initial_state)))
+        
+
         return pipeline
 
     @staticmethod
-    def from_proto(pipeline_proto: Pipeline, serialiser, events, storage):
+    def from_proto(pipeline_proto: Pipeline, serialiser, events):
         def _from_proto(entry):
             return Task.from_proto(entry) if entry.HasField('task_entry') else Group.from_proto(entry)
 
-        pipeline = [_from_proto(proto) for proto in pipeline_proto.entries]
-
-        return _Pipeline(pipeline, serialiser=serialiser, events=events, storage=storage)
+        return _Pipeline(
+            pipeline=[_from_proto(proto) for proto in pipeline_proto.entries], 
+            inline=pipeline_proto.inline, 
+            initial_args=pipeline_proto.initial_args,
+            initial_state=pipeline_proto.initial_state,\
+            serialiser=serialiser, events=events)
 
     async def handle_message(self, context: TaskContext, message: Union[TaskRequest, TaskResult, TaskException], state: Any=None) -> bool:
         handled = False
@@ -110,7 +128,7 @@ class _Pipeline(object):
         await self.events.notify_pipeline_status_changed(context, context.pipeline_state.pipeline, context.pipeline_state.status.value)
 
         try:
-            await self._submitter.unpause_tasks(context)
+            self._submitter.unpause_tasks(context)
 
             # tell any child pipelines to resume
             for child_pipeline in context.pipeline_state.child_pipelines:
