@@ -1,12 +1,27 @@
+from re import L
 from statefun_tasks.types import Task, Group, RetryPolicy, ProtobufSerialisable
 from statefun_tasks.utils import _gen_id, _is_tuple
 from statefun_tasks.pipeline import _Pipeline
 from statefun_tasks.messages_pb2 import TaskRequest, Pipeline
-
+from statefun_tasks.builtin import builtin
 from typing import Iterable
+import math
 
 
-def in_parallel(entries: list, max_parallelism=None):
+def in_parallel(entries: list, max_parallelism=None, num_stages:int = 1):
+    if num_stages > 1:
+        # split up a group such [[1,2,3,4,5,6]] into inline pipelines each with a subset of the group
+        # i.e. [[p[1,2], p[3,4], p[5,6]]] followed by a flatten to allow for better distribution
+        # of a parallelism over multiple workers
+        chunk_size = max(math.ceil(len(entries) / num_stages), 1)
+        per_stage_max_parallelism = None if max_parallelism is None else max(int(max_parallelism / num_stages), 1)
+        stages = [entries[i:i + chunk_size] for i in range(0, len(entries), chunk_size)]
+
+        if len(stages) > 1:
+            per_stage_pipeline = [PipelineBuilder().append_group(stage, max_parallelism=per_stage_max_parallelism) for stage in stages]
+            group = [PipelineBuilder().append(builtin.run_pipeline.to_task(args=pipeline.inline())) for pipeline in per_stage_pipeline] 
+            return PipelineBuilder().append_group(group, max_parallelism=max_parallelism).continue_with(builtin.flatten_results)
+
     return PipelineBuilder().append_group(entries, max_parallelism)
 
 
@@ -16,10 +31,12 @@ class PipelineBuilder(ProtobufSerialisable):
 
     :param optional pipeline: list of initial pipeline entries e.g. from another builder
     """
-
     def __init__(self, pipeline: list = None):
         self._pipeline = [] if pipeline is None else pipeline
         self._builder_id = _gen_id()
+        self._inline = False
+        self._initial_args = None
+        self._initial_state = None
 
     @property
     def id(self):
@@ -32,6 +49,13 @@ class PipelineBuilder(ProtobufSerialisable):
     def id(self, value):
         self._builder_id = value
 
+    @property
+    def is_inline(self):
+        """
+        Returns true if the pipeline is inline or not
+        """
+        return self._inline
+
     def append(self, task: Task) -> 'PipelineBuilder':
         """
         Appends a single task onto this pipeline
@@ -176,20 +200,16 @@ class PipelineBuilder(ProtobufSerialisable):
 
     def to_task_request(self, serialiser) -> TaskRequest:
         """
-        Serialises the pipeline as a TaskRequest with a task type of '__builtins.run_pipeline'
+        Serialises the pipeline as a TaskRequest with a task type of builtin.run_pipeline
 
         :param serialiser: the serialiser to use such as DefaultSerialiser
         :return: TaskRequest protobuf message
         """
         task_id = self._builder_id
-        task_type = '__builtins.run_pipeline'
+        task_type = builtin.run_pipeline.task_name
         args = self.validate().to_proto(serialiser)
         kwargs = {}
 
-        # send a single argument by itself instead of wrapped inside a tuple
-        if _is_tuple(args) and len(args) == 1:
-            args = args[0]
-
         task_request = TaskRequest(id=task_id, type=task_type, uid=_gen_id())
         args_and_kwargs = serialiser.serialise_args_and_kwargs(args, kwargs)
         serialiser.serialise_request(task_request, args_and_kwargs)
@@ -223,7 +243,51 @@ class PipelineBuilder(ProtobufSerialisable):
         :return: a Flink Tasks pipeline
         """
         self.validate()
-        return _Pipeline(self._pipeline, events=events, serialiser=serialiser, is_fruitful=is_fruitful)
+
+        return _Pipeline(self._pipeline, 
+            inline=self._inline, 
+            initial_args=self._initial_args,
+            initial_state=self._initial_state,
+            is_fruitful=is_fruitful,
+            serialiser=serialiser,
+            events=events)
+
+    def set_task_defaults(self, default_namespace, default_worker_name) ->  'PipelineBuilder':
+        """
+        Sets defaults on task entries if they are not set
+        :return: the builder
+        """
+        for task in self._get_tasks():
+            if task.namespace == '':
+                task.namespace = default_namespace
+            if task.worker_name == '':
+                task.worker_name = default_worker_name
+
+        return self
+
+    def with_initial(self, args=Ellipsis, state=Ellipsis) ->  'PipelineBuilder':
+        """
+        Optionally sets the initial args and state to be passed to the initial tasks(s) in this pipeline
+        :param option args: arguments
+        :param option state: state
+        :return: the builder
+        """
+        if args != Ellipsis:
+            self._initial_args = args
+
+        if state != Ellipsis:
+            self._initial_state = state
+            
+        return self
+
+    def inline(self, is_inline=True) ->  'PipelineBuilder':
+        """
+        Marks the pipeline as being inline (or not).  By default pipelines are not inline.
+        Inline pipelines accept inputs from and share state with their parent task.
+        :return: the builder
+        """
+        self._inline = is_inline
+        return self
 
     def validate(self) -> 'PipelineBuilder':
         """
@@ -263,7 +327,7 @@ class PipelineBuilder(ProtobufSerialisable):
         :param serialiser: the serialiser to use such as DefaultSerialiser
         :return: Pipeline protobuf message
         """
-        return _Pipeline(self._pipeline, serialiser=serialiser).to_proto()
+        return self.to_pipeline(serialiser=serialiser).to_proto()
 
     @staticmethod
     def from_proto(pipeline_proto: Pipeline) -> 'PipelineBuilder':
@@ -281,7 +345,21 @@ class PipelineBuilder(ProtobufSerialisable):
             elif proto.HasField('group_entry'):
                 pipeline.append(Group.from_proto(proto))
 
-        return PipelineBuilder(pipeline)
+        builder = PipelineBuilder(pipeline)
+
+        if pipeline_proto.inline:
+            builder = builder.inline()
+
+        args, state = None, None
+        if pipeline_proto.HasField('initial_args'):
+            args = pipeline_proto.initial_args
+
+        if pipeline_proto.HasField('initial_state'):
+            state = pipeline_proto.initial_state
+
+        builder = builder.with_initial(args, state)
+
+        return builder
 
     def get_tasks(self) -> list:
         """
