@@ -1,6 +1,7 @@
 from statefun_tasks.context import TaskContext
 from statefun_tasks.pipeline_impl.handlers import PipelineMessageHandler
 from statefun_tasks.messages_pb2 import TaskRequest, TaskResult, TaskException, TaskStatus, PipelineState, ChildPipeline, TaskInfo
+from statefun_tasks.protobuf import unpack_any
 from statefun_tasks.serialisation import pack_any
 from statefun_tasks.utils import _gen_id
 from statefun_tasks.type_helpers import _create_task_result
@@ -63,6 +64,18 @@ class BeginPipelineHandler(PipelineMessageHandler):
             # else get initial tasks(s) to call - might be single start of chain task or a group of tasks to call in parallel
             tasks, max_parallelism, slice = self.graph.get_initial_tasks()
 
+            # send initial state to each initial task if we have some
+            if context.pipeline_state.pipeline.HasField('initial_state'):                
+                task_state = unpack_any(context.pipeline_state.pipeline.initial_state, known_proto_types=[])
+            else:
+                task_state = None
+
+            # send initial arguments to each initial task if we have them
+            if context.pipeline_state.pipeline.HasField('initial_args'):
+                initial_args = unpack_any(context.pipeline_state.pipeline.initial_args, known_proto_types=[])
+            else:
+                initial_args = None
+
             # if we skipped over empty group(s) then make sure we pass empty array to next task (result of in_parallel([]) is intuitively [])
             if slice > 0:
                 for task in tasks:
@@ -70,7 +83,7 @@ class BeginPipelineHandler(PipelineMessageHandler):
                     task.request = self._serialiser.serialise_args_and_kwargs(([]), kwargs)
 
             # split into tasks to call now and those to defer if max parallelism is exceeded
-            await self.submitter.submit_tasks(context, tasks, max_parallelism=max_parallelism)
+            self.submitter.submit_tasks(context, tasks, task_state=task_state, max_parallelism=max_parallelism, initial_args=initial_args)
 
             # break
             return False, message
